import * as ts from 'typescript';
import { OutputContext, Transpiler } from './main';
export declare type ClassLike = ts.ClassDeclaration | ts.InterfaceDeclaration;
export declare type NamedDeclaration = ClassLike | ts.PropertyDeclaration | ts.VariableDeclaration | ts.MethodDeclaration | ts.ModuleDeclaration | ts.FunctionDeclaration;
export declare type Set = {
    [s: string]: boolean;
};
export declare function ident(n: ts.Node): string;
export declare function isFunctionTypedefLikeInterface(ifDecl: ts.InterfaceDeclaration): boolean;
export declare function getDeclaration(type: ts.Type): ts.Declaration;
export declare function isExtendsClause(heritageClause: ts.HeritageClause): boolean;
export declare function isConstructor(n: ts.Node): boolean;
export declare function isStatic(n: ts.Node): boolean;
export declare function isCallableType(type: ts.TypeNode, tc: ts.TypeChecker): boolean;
export declare function isFunctionType(type: ts.TypeNode, tc: ts.TypeChecker): boolean;
export declare function isTypeNode(node: ts.Node): boolean;
export declare function isCallable(decl: ClassLike): boolean;
export declare function copyLocation(src: ts.TextRange, dest: ts.TextRange): void;
export declare function arrayFindPolyfill<T>(nodeArray: ts.NodeArray<T>, predicate: (node: T) => boolean): T;
export declare function getAncestor(n: ts.Node, kind: ts.SyntaxKind): ts.Node;
export declare function getEnclosingClass(n: ts.Node): ClassLike;
export declare function isConstCall(node: ts.CallExpression): boolean;
export declare function isInsideConstExpr(node: ts.Node): boolean;
export declare class TranspilerBase {
    protected transpiler: Transpiler;
    private idCounter;
    constructor(transpiler: Transpiler);
    visit(n: ts.Node): void;
    pushContext(context: OutputContext): void;
    popContext(): void;
    emit(s: string): void;
    emitNoSpace(s: string): void;
    enterCodeComment(): void;
    exitCodeComment(): void;
    emitImport(toEmit: string): void;
    reportError(n: ts.Node, message: string): void;
    visitNode(n: ts.Node): boolean;
    visitEach(nodes: ts.Node[]): void;
    visitEachIfPresent(nodes?: ts.Node[]): void;
    visitList(nodes: ts.Node[], separator?: string): void;
    uniqueId(name: string): string;
    assert(c: ts.Node, condition: boolean, reason: string): void;
    getAncestor(n: ts.Node, kind: ts.SyntaxKind): ts.Node;
    hasAncestor(n: ts.Node, kind: ts.SyntaxKind): boolean;
    hasAnnotation(decorators: ts.NodeArray<ts.Decorator>, name: string): boolean;
    hasFlag(n: {
        flags: number;
    }, flag: ts.NodeFlags): boolean;
    getRelativeFileName(fileName: string): string;
    maybeVisitTypeArguments(n: {
        typeArguments?: ts.NodeArray<ts.TypeNode>;
    }): void;
    visitParameters(parameters: ts.ParameterDeclaration[]): void;
}
