import * as ts from 'typescript';
import * as base from './base';
import { FacadeConverter } from './facade_converter';
import { Transpiler } from './main';
export declare function isFunctionLikeProperty(decl: ts.PropertyDeclaration | ts.ParameterDeclaration, tc: ts.TypeChecker): boolean;
export default class DeclarationTranspiler extends base.TranspilerBase {
    private fc;
    private enforceUnderscoreConventions;
    private promoteFunctionLikeMembers;
    private tc;
    private moduleStack;
    private extendsClass;
    static NUM_FAKE_REST_PARAMETERS: number;
    setTypeChecker(tc: ts.TypeChecker): void;
    setFacadeConverter(fc: FacadeConverter): void;
    getJsPath(node: ts.Node): string;
    private isAnonymousInterface(node);
    maybeEmitJsAnnotation(node: ts.Node): void;
    /**
     * Emit fake constructors to placate the Dart Analyzer for JS Interop classes.
     */
    maybeEmitFakeConstructors(decl: base.ClassLike): void;
    private visitName(name);
    private notSimpleBagOfProperties(type);
    /**
     * Returns whether all members of the class and all base classes
     */
    hasOnlyProperties(decl: ts.InterfaceDeclaration, outProperties: ts.PropertyDeclaration[]): boolean;
    visitClassBody(decl: base.ClassLike): void;
    visitMergingOverloads(members: Array<ts.Node>): void;
    constructor(tr: Transpiler, fc: FacadeConverter, enforceUnderscoreConventions: boolean, promoteFunctionLikeMembers: boolean);
    visitNode(node: ts.Node): boolean;
    private visitFunctionLike(fn, accessor?);
    /**
     * Visit a property declaration.
     * In the special case of property parameters in a constructor, we also allow
     * a parameter to be emitted as a property.
     * We have to emit properties as getter setter pairs as Dart does not support
     * external fields.
     * In the special case of property parameters in a constructor, we also allow a parameter to be
     * emitted as a property.
     */
    private visitProperty(decl, isParameter?);
    private visitClassLike(keyword, decl);
    private visitDeclarationMetadata(decl);
    /**
     * Handles a function typedef-like interface, i.e. an interface that only declares a single
     * call signature, by translating to a Dart `typedef`.
     */
    private visitFunctionTypedefInterface(name, signature, typeParameters);
}
