import * as ts from 'typescript';
import * as base from './base';
import { Transpiler } from './main';
export declare function fixupIdentifierName(text: string): string;
export declare class DartLibrary {
    private fileName;
    constructor(fileName: string);
    /**
     * @returns {boolean} whether the name was added.
     */
    addName(name: string): boolean;
    private usedNames;
}
export declare class NameRewriter {
    private dartTypes;
    private libraries;
    private computeName(node);
    lookupName(node: base.NamedDeclaration, context: ts.Node): string;
}
export declare class FacadeConverter extends base.TranspilerBase {
    private nameRewriter;
    private useHtml;
    private tc;
    static DART_RESERVED_WORDS: string[];
    static DART_OTHER_KEYWORDS: string[];
    private candidateTypes;
    private typingsRootRegex;
    private genericMethodDeclDepth;
    constructor(transpiler: Transpiler, typingsRoot: string, nameRewriter: NameRewriter, useHtml: boolean);
    private extractPropertyNames(m, candidates);
    setTypeChecker(tc: ts.TypeChecker): void;
    pushTypeParameterNames(n: ts.FunctionLikeDeclaration): void;
    popTypeParameterNames(n: ts.FunctionLikeDeclaration): void;
    resolvePropertyTypes(tn: ts.TypeNode): ts.Map<ts.PropertyDeclaration>;
    /**
     * The Dart analyzer has a syntax extension that uses comments to emulate
     * generic methods in Dart. We work around this and keep track of which type
     * names in the current scope need to be emitted in comments.
     *
     * TODO(jacobr): Remove this once all Dart implementations support generic
     * methods.
     */
    private isGenericMethodTypeParameterName(name);
    generateTypeList(types: ts.TypeNode[], seperator?: string): string;
    maybeGenerateTypeArguments(n: {
        typeArguments?: ts.NodeArray<ts.TypeNode>;
    }): string;
    generateDartTypeName(node: ts.TypeNode): string;
    visitTypeName(typeName: ts.EntityName): void;
    getSymbolDeclaration(symbol: ts.Symbol, n?: ts.Node): ts.Declaration;
    generateDartName(identifier: ts.EntityName): string;
    /**
     * Returns null if declaration cannot be found or is not valid in Dart.
     */
    getDeclaration(identifier: ts.EntityName): ts.Declaration;
    /**
     * Returns a custom Dart type name or null if the type isn't a custom Dart
     * type.
     */
    lookupCustomDartTypeName(identifier: ts.EntityName): {
        name?: string;
        comment?: string;
        keep?: boolean;
    };
    /**
     * This method works around the lack of Dart support for union types
     * generating a valid Dart type that satisfies all the types passed in.
     */
    toSimpleDartType(types: Array<ts.TypeNode>): ts.TypeNode;
    toTypeNode(type: ts.Type): ts.TypeNode;
    createEntityName(symbol: ts.Symbol): ts.EntityName;
    safeGetBaseTypes(type: ts.InterfaceType): ts.ObjectType[];
    checkTypeSubtypeOf(source: ts.Type, target: ts.Type): boolean;
    commonSupertype(nodeA: ts.TypeNode, nodeB: ts.TypeNode): ts.TypeNode;
    getCommonSupertype(a: ts.Type, b: ts.Type): ts.Type;
    private getFileAndName(n, originalSymbol);
}
