"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var ts = require('typescript');
var base = require('./base');
var dart_libraries_for_browser_types_1 = require('./dart_libraries_for_browser_types');
var merge_1 = require('./merge');
var FACADE_DEBUG = false;
var FACADE_NODE_MODULES_PREFIX = /^(\.\.\/)*node_modules\//;
// These constants must be kept in sync with package:func/func.dart which
// provides a cannonical set of typedefs defining commonly used function types
// to simplify specifying function types in Dart.
var MAX_DART_FUNC_ACTION_PARAMETERS = 4;
var MAX_DART_FUNC_ACTION_PARAMETERS_OPTIONAL = 1;
/**
 * Prefix to add to a variable name that leaves the JS name referenced
 * unchanged.
 */
var DART_RESERVED_NAME_PREFIX = 'JS$';
function fixupIdentifierName(text) {
    return (FacadeConverter.DART_RESERVED_WORDS.indexOf(text) !== -1 ||
        FacadeConverter.DART_OTHER_KEYWORDS.indexOf(text) !== -1 || text[0] === '_') ?
        DART_RESERVED_NAME_PREFIX + text :
        text;
}
exports.fixupIdentifierName = fixupIdentifierName;
function numOptionalParameters(parameters) {
    for (var i = 0; i < parameters.length; ++i) {
        if (parameters[i].questionToken)
            return parameters.length - i;
    }
    return 0;
}
function hasVarArgs(parameters) {
    for (var i = 0; i < parameters.length; ++i) {
        if (parameters[i].dotDotDotToken)
            return true;
    }
    return false;
}
/**
 * Generates the JavaScript expression required to reference the node
 * from the global context. InterfaceDeclarations do not technically correspond
 * to actual JavaScript objects but we still generate a reference path for them
 * so that we have a guaranteed unique name.
 *
 * Example JS path:
 * module m1 {
 *   module m2 {
 *     class foo { }
 *   }
 * }
 * Path: m1.m2.foo
 */
function fullJsPath(node) {
    var parts = [base.ident(node.name)];
    var p = node.parent;
    while (p != null) {
        var kind = p.kind;
        if (kind === ts.SyntaxKind.ModuleDeclaration || kind === ts.SyntaxKind.InterfaceDeclaration ||
            kind === ts.SyntaxKind.ClassDeclaration) {
            parts.unshift(base.ident(p.name));
        }
        p = p.parent;
    }
    return parts.join('.');
}
var DartNameRecord = (function () {
    function DartNameRecord(node, name, library) {
        this.node = node;
        this.library = library;
        this.name = name;
    }
    return DartNameRecord;
}());
var DartLibrary = (function () {
    function DartLibrary(fileName) {
        this.fileName = fileName;
        this.usedNames = {};
    }
    /**
     * @returns {boolean} whether the name was added.
     */
    DartLibrary.prototype.addName = function (name) {
        if (Object.prototype.hasOwnProperty.call(this.usedNames, name)) {
            return false;
        }
        this.usedNames[name] = true;
        return true;
    };
    return DartLibrary;
}());
exports.DartLibrary = DartLibrary;
var NameRewriter = (function () {
    function NameRewriter() {
        this.dartTypes = {};
        // TODO(jacobr): use libraries to track what library imports need to be
        // emitted. Complex cases such as ts.SyntaxKind.TypeReference
        // where emitting the type could require emitting imports to libraries not
        // specified in the imports listed in TypeScript. Additionally, d.ts files
        // can contain multiple modules and for readability we may want an option
        // to emit each of those modules as a separate Dart library. That would also
        // require tracking what external libraries are referenced.
        this.libraries = {};
    }
    NameRewriter.prototype.computeName = function (node) {
        var fullPath = fullJsPath(node);
        if (Object.prototype.hasOwnProperty.call(this.dartTypes, fullPath)) {
            return this.dartTypes[fullPath];
        }
        var sourceFile = base.getAncestor(node, ts.SyntaxKind.SourceFile);
        var fileName = sourceFile.fileName;
        var library;
        if (Object.prototype.hasOwnProperty.call(this.libraries, fileName)) {
            library = this.libraries[fileName];
        }
        else {
            library = new DartLibrary(fileName);
            this.libraries[fileName] = library;
        }
        var parts = fullPath.split('.');
        for (var i_1 = parts.length - 1; i_1 >= 0; i_1--) {
            // Find a unique name by including more of the module hierarchy in the
            // name. This is an arbitrary but hopefully unsurprising scheme to
            // generate unique names. There may be classes or members with conflicting
            // names due to a single d.ts file containing multiple modules.
            // TODO(jacobr): we should suppress this behavior outside of JS Interop
            // mode and instead generate a compile error if there are conflicting
            // names.
            var candidateName = fixupIdentifierName(parts.slice(i_1).join('_'));
            if (library.addName(candidateName)) {
                // Able to add name to library.
                var ret = new DartNameRecord(node, candidateName, library);
                this.dartTypes[fullPath] = ret;
                return ret;
            }
        }
        // Usually the module name prefixes should be sufficient to disambiguate
        // names but sometimes we need to add a numeric prefix as well to
        // disambiguate. We could alternately append the full module prefix as well
        // to make the name choice completely unsurprising albeit even uglier.
        // This case should be very rarely hit.
        var i = 2;
        while (true) {
            var candidateName = parts[parts.length - 1] + i;
            if (library.addName(candidateName)) {
                // Able to add name to library.
                var ret = new DartNameRecord(node, candidateName, library);
                this.dartTypes[fullPath] = ret;
                return ret;
            }
            i++;
        }
    };
    NameRewriter.prototype.lookupName = function (node, context) { return this.computeName(node).name; };
    return NameRewriter;
}());
exports.NameRewriter = NameRewriter;
var FacadeConverter = (function (_super) {
    __extends(FacadeConverter, _super);
    function FacadeConverter(transpiler, typingsRoot, nameRewriter, useHtml) {
        var _this = this;
        if (typingsRoot === void 0) { typingsRoot = ''; }
        _super.call(this, transpiler);
        this.nameRewriter = nameRewriter;
        this.useHtml = useHtml;
        this.candidateTypes = {};
        this.genericMethodDeclDepth = 0;
        if (useHtml) {
            this.extractPropertyNames(dart_libraries_for_browser_types_1.TS_TO_DART_TYPENAMES, this.candidateTypes);
            Object.keys(dart_libraries_for_browser_types_1.DART_LIBRARIES_FOR_BROWSER_TYPES)
                .forEach(function (propName) { return _this.candidateTypes[propName] = true; });
        }
        else {
            this.extractPropertyNames(dart_libraries_for_browser_types_1.TS_TO_DART_TYPENAMES, this.candidateTypes);
        }
        this.typingsRootRegex = new RegExp('^' + typingsRoot.replace('.', '\\.'));
    }
    FacadeConverter.prototype.extractPropertyNames = function (m, candidates) {
        for (var _i = 0, _a = Object.keys(m); _i < _a.length; _i++) {
            var fileName = _a[_i];
            var file = m[fileName];
            if (file === undefined) {
                return;
            }
            Object.keys(file)
                .map(function (propName) { return propName.substring(propName.lastIndexOf('.') + 1); })
                .forEach(function (propName) { return candidates[propName] = true; });
        }
    };
    FacadeConverter.prototype.setTypeChecker = function (tc) { this.tc = tc; };
    FacadeConverter.prototype.pushTypeParameterNames = function (n) {
        if (!n.typeParameters)
            return;
        this.genericMethodDeclDepth++;
    };
    FacadeConverter.prototype.popTypeParameterNames = function (n) {
        if (!n.typeParameters)
            return;
        this.genericMethodDeclDepth--;
    };
    FacadeConverter.prototype.resolvePropertyTypes = function (tn) {
        var res = {};
        if (!tn || !this.tc)
            return res;
        var t = this.tc.getTypeAtLocation(tn);
        for (var _i = 0, _a = this.tc.getPropertiesOfType(t); _i < _a.length; _i++) {
            var sym = _a[_i];
            var decl = sym.valueDeclaration || (sym.declarations && sym.declarations[0]);
            if (decl.kind !== ts.SyntaxKind.PropertyDeclaration &&
                decl.kind !== ts.SyntaxKind.PropertySignature) {
                var msg = this.tc.getFullyQualifiedName(sym) +
                    ' used for named parameter definition must be a property';
                this.reportError(decl, msg);
                continue;
            }
            res[sym.name] = decl;
        }
        return res;
    };
    /**
     * The Dart analyzer has a syntax extension that uses comments to emulate
     * generic methods in Dart. We work around this and keep track of which type
     * names in the current scope need to be emitted in comments.
     *
     * TODO(jacobr): Remove this once all Dart implementations support generic
     * methods.
     */
    FacadeConverter.prototype.isGenericMethodTypeParameterName = function (name) {
        // Avoid checking this unless needed.
        if (this.genericMethodDeclDepth === 0 || !this.tc)
            return false;
        // Check if the type of the name is a TypeParameter.
        var t = this.tc.getTypeAtLocation(name);
        if (!t || (t.flags & ts.TypeFlags.TypeParameter) === 0)
            return false;
        // Check if the symbol we're looking at is the type parameter.
        var symbol = this.tc.getSymbolAtLocation(name);
        if (symbol !== t.symbol)
            return false;
        // Check that the Type Parameter has been declared by a function declaration.
        return symbol.declarations.some(function (d) { return d.parent.kind === ts.SyntaxKind.FunctionDeclaration; });
    };
    FacadeConverter.prototype.generateTypeList = function (types, seperator) {
        if (seperator === void 0) { seperator = ','; }
        return types.map(this.generateDartTypeName.bind(this)).join(seperator);
    };
    FacadeConverter.prototype.maybeGenerateTypeArguments = function (n) {
        if (!n.typeArguments)
            return '';
        return '<' + this.generateTypeList(n.typeArguments) + '>';
    };
    FacadeConverter.prototype.generateDartTypeName = function (node) {
        var name;
        var comment;
        if (!node) {
            return 'dynamic';
        }
        switch (node.kind) {
            case ts.SyntaxKind.TypeQuery:
                var query = node;
                name = 'dynamic';
                name += '/* Dart does not support TypeQuery: typeof ' + base.ident(query.exprName) + ' */';
                break;
            case ts.SyntaxKind.LastTypeNode:
                var type = node.type;
                if (!type) {
                    // This case occurs for String literal types
                    comment = node.getText();
                    // TODO(jacobr): find a better way to detect string literal types.
                    name = comment[0] === '"' ? 'String' : 'dynamic';
                    break;
                }
                return this.generateDartTypeName(type);
            case ts.SyntaxKind.TypePredicate:
                return this.generateDartTypeName(node.type);
            case ts.SyntaxKind.TupleType:
                var tuple = node;
                name = 'List<';
                var mergedType_1 = new merge_1.MergedType(this);
                tuple.elementTypes.forEach(function (t) { return mergedType_1.merge(t); });
                name += this.generateDartTypeName(mergedType_1.toTypeNode());
                name += '>';
                comment = 'Tuple<' + this.generateTypeList(tuple.elementTypes) + '>';
                break;
            case ts.SyntaxKind.UnionType:
                var union = node;
                // TODO(jacobr): this isn't fundamentally JS Interop specific but we
                // choose to be more aggressive at finding a useful value for the
                // union when in JS Interop mode while otherwise we expect that union
                // types will not be used extensively.
                var simpleType = this.toSimpleDartType(union.types);
                if (simpleType) {
                    name = this.generateDartTypeName(simpleType);
                }
                else {
                    name = 'dynamic';
                }
                var types = union.types;
                comment = this.generateTypeList(types, '|');
                break;
            case ts.SyntaxKind.TypePredicate:
                return this.generateDartTypeName(node.type);
            case ts.SyntaxKind.TypeReference:
                var typeRef = node;
                name = this.generateDartName(typeRef.typeName) + this.maybeGenerateTypeArguments(typeRef);
                break;
            case ts.SyntaxKind.TypeLiteral:
                var members = node.members;
                if (members.length === 1 && members[0].kind === ts.SyntaxKind.IndexSignature) {
                    var indexSig = (members[0]);
                    if (indexSig.parameters.length > 1) {
                        this.reportError(indexSig, 'Expected an index signature to have a single parameter');
                    }
                    // Unfortunately for JS interop, we cannot treat JS Objects as Dart
                    // Map objects. We could treat them as JSMap<indexSig.type>
                    // if we define a base JSMap type that is Map like but not actually
                    // a map.
                    name = 'dynamic';
                    comment = 'Map<' + this.generateDartTypeName(indexSig.parameters[0].type) + ',' +
                        this.generateDartTypeName(indexSig.type) + '>';
                }
                else {
                    name = 'dynamic';
                    comment = node.getText();
                }
                break;
            case ts.SyntaxKind.FunctionType:
                var callSignature = node;
                var parameters = callSignature.parameters;
                // Use a function signature from package:func where possible.
                var numOptional = numOptionalParameters(parameters);
                var isVoid = callSignature.type && callSignature.type.kind === ts.SyntaxKind.VoidKeyword;
                if (parameters.length <= MAX_DART_FUNC_ACTION_PARAMETERS &&
                    numOptional <= MAX_DART_FUNC_ACTION_PARAMETERS_OPTIONAL && !hasVarArgs(parameters)) {
                    this.emitImport('package:func/func.dart');
                    var typeDefName = (isVoid) ? 'VoidFunc' : 'Func';
                    typeDefName += parameters.length;
                    if (numOptional > 0) {
                        typeDefName += 'Opt' + numOptional;
                    }
                    name = typeDefName;
                    var numArgs = parameters.length + (isVoid ? 0 : 1);
                    if (numArgs > 0) {
                        name += '<';
                    }
                    var isFirst = true;
                    for (var i = 0; i < parameters.length; ++i) {
                        if (isFirst) {
                            isFirst = false;
                        }
                        else {
                            name += ', ';
                        }
                        name += this.generateDartTypeName(parameters[i].type);
                    }
                    if (!isVoid) {
                        if (!isFirst) {
                            name += ', ';
                        }
                        name += this.generateDartTypeName(callSignature.type);
                    }
                    if (numArgs > 0) {
                        name += '>';
                    }
                }
                else {
                    name = 'Function';
                    if (node.getSourceFile()) {
                        comment = node.getText();
                    }
                }
                break;
            case ts.SyntaxKind.ArrayType:
                name = 'List' +
                    '<' + this.generateDartTypeName(node.elementType) + '>';
                break;
            case ts.SyntaxKind.NumberKeyword:
                name = 'num';
                break;
            case ts.SyntaxKind.StringLiteral:
            case ts.SyntaxKind.StringKeyword:
                name = 'String';
                break;
            case ts.SyntaxKind.VoidKeyword:
                name = 'void';
                break;
            case ts.SyntaxKind.BooleanKeyword:
                name = 'bool';
                break;
            case ts.SyntaxKind.AnyKeyword:
                name = 'dynamic';
                break;
            default:
                this.reportError(node, 'Unexpected TypeNode kind');
        }
        if (name == null) {
            name = 'XXX NULLNAME';
        }
        name = name.trim();
        if (comment) {
            name = name + '/*' + comment + '*/';
        }
        return name;
    };
    FacadeConverter.prototype.visitTypeName = function (typeName) {
        if (typeName.kind !== ts.SyntaxKind.Identifier) {
            this.visit(typeName);
            return;
        }
        var ident = base.ident(typeName);
        if (this.isGenericMethodTypeParameterName(typeName)) {
            // DDC generic methods hack - all names that are type parameters to generic methods have to be
            // emitted in comments.
            this.emit('dynamic/*=');
            this.emit(ident);
            this.emit('*/');
            return;
        }
        var custom = this.lookupCustomDartTypeName(typeName);
        if (custom) {
            if (custom.comment) {
                this.emit('/*=' + custom.comment + '*/');
            }
            this.emit(custom.name);
        }
        else {
            this.visit(typeName);
        }
    };
    FacadeConverter.prototype.getSymbolDeclaration = function (symbol, n) {
        if (!symbol)
            return null;
        var decl = symbol.valueDeclaration;
        if (!decl) {
            // In the case of a pure declaration with no assignment, there is no value declared.
            // Just grab the first declaration, hoping it is declared once.
            if (!symbol.declarations || symbol.declarations.length === 0) {
                this.reportError(n, 'no declarations for symbol ' + symbol.name);
                return;
            }
            decl = symbol.declarations[0];
        }
        return decl;
    };
    FacadeConverter.prototype.generateDartName = function (identifier) {
        var ret = this.lookupCustomDartTypeName(identifier);
        if (ret) {
            // TODO(jacobr): reduce the number of times we write this pattern.
            if (ret.comment) {
                return ret.name + '/*=' + ret.comment + '*/';
            }
            else {
                return ret.name;
            }
        }
        // TODO(jacobr): handle library import prefixes better. This generally works
        // but is somewhat fragile.
        return base.ident(identifier);
    };
    /**
     * Returns null if declaration cannot be found or is not valid in Dart.
     */
    FacadeConverter.prototype.getDeclaration = function (identifier) {
        var symbol;
        if (!this.tc)
            return null;
        symbol = this.tc.getSymbolAtLocation(identifier);
        var declaration = this.getSymbolDeclaration(symbol, identifier);
        if (symbol && symbol.flags & ts.SymbolFlags.TypeParameter) {
            var kind = declaration.parent.kind;
            // Only kinds of TypeParameters supported by Dart.
            if (kind !== ts.SyntaxKind.ClassDeclaration && kind !== ts.SyntaxKind.InterfaceDeclaration) {
                return null;
            }
        }
        return declaration;
    };
    /**
     * Returns a custom Dart type name or null if the type isn't a custom Dart
     * type.
     */
    FacadeConverter.prototype.lookupCustomDartTypeName = function (identifier) {
        var ident = base.ident(identifier);
        var symbol;
        if (!this.tc)
            return null;
        symbol = this.tc.getSymbolAtLocation(identifier);
        var declaration = this.getSymbolDeclaration(symbol, identifier);
        if (symbol && symbol.flags & ts.SymbolFlags.TypeParameter) {
            var kind = declaration.parent.kind;
            // Only kinds of TypeParameters supported by Dart.
            if (kind !== ts.SyntaxKind.ClassDeclaration && kind !== ts.SyntaxKind.InterfaceDeclaration) {
                return { name: 'dynamic', comment: ident };
            }
        }
        if (this.candidateTypes.hasOwnProperty(ident)) {
            if (!symbol) {
                return null;
            }
            var fileAndName = this.getFileAndName(identifier, symbol);
            if (fileAndName) {
                var fileSubs = dart_libraries_for_browser_types_1.TS_TO_DART_TYPENAMES[fileAndName.fileName];
                if (fileSubs) {
                    var dartBrowserType = dart_libraries_for_browser_types_1.DART_LIBRARIES_FOR_BROWSER_TYPES.hasOwnProperty(fileAndName.qname);
                    if (dartBrowserType) {
                        this.emitImport(dart_libraries_for_browser_types_1.DART_LIBRARIES_FOR_BROWSER_TYPES[fileAndName.qname]);
                    }
                    if (fileSubs.hasOwnProperty(fileAndName.qname)) {
                        return { name: fileSubs[fileAndName.qname] };
                    }
                    if (dartBrowserType) {
                        // Not a rename but has a dart core libraries definition.
                        return { name: fileAndName.qname };
                    }
                }
            }
        }
        if (symbol) {
            if (symbol.flags & ts.SymbolFlags.Enum) {
                // We can't treat JavaScript enums as Dart enums in this case.
                return { name: 'num', comment: ident };
            }
            // TODO(jacobr): we could choose to only support type alais declarations
            // for JS interop but it seems handling type alaises is generally helpful
            // without much risk of generating confusing Dart code.
            if (declaration.kind === ts.SyntaxKind.TypeAliasDeclaration) {
                var alias = declaration;
                if (alias.typeParameters) {
                    // We can handle this case but currently do not.
                    this.reportError(declaration, 'Type parameters for type alaises are not supported');
                }
                return { name: this.generateDartTypeName(alias.type) };
            }
            var kind = declaration.kind;
            if (kind === ts.SyntaxKind.ClassDeclaration || kind === ts.SyntaxKind.InterfaceDeclaration ||
                kind === ts.SyntaxKind.VariableDeclaration ||
                kind === ts.SyntaxKind.PropertyDeclaration ||
                kind === ts.SyntaxKind.FunctionDeclaration) {
                var name_1 = this.nameRewriter.lookupName(declaration, identifier);
                if (kind === ts.SyntaxKind.InterfaceDeclaration &&
                    base.isFunctionTypedefLikeInterface(declaration) &&
                    base.getAncestor(identifier, ts.SyntaxKind.HeritageClause)) {
                    // TODO(jacobr): we need to specify a specific call method for this
                    // case if we want to get the most from Dart type checking.
                    return { name: 'Function', comment: name_1 };
                }
                return { name: name_1, keep: true };
            }
        }
        return null;
    };
    // TODO(jacobr): performance of this method could easily be optimized.
    /**
     * This method works around the lack of Dart support for union types
     * generating a valid Dart type that satisfies all the types passed in.
     */
    FacadeConverter.prototype.toSimpleDartType = function (types) {
        // We use MergeType to ensure that we have already deduped types that are
        // equivalent even if they aren't obviously identical.
        // MergedType will also follow typed aliases, etc which allows us to avoid
        // including that logic here as well.
        var mergedType = new merge_1.MergedType(this);
        types.forEach(function (type) { mergedType.merge(type); });
        var merged = mergedType.toTypeNode();
        if (merged.kind === ts.SyntaxKind.UnionType) {
            // For union types find a Dart type that satisfies all the types.
            types = merged.types;
            /**
             * Generate a common base type for an array of types.
             * The implemented is currently incomplete often returning null when there
             * might really be a valid common base type.
             */
            var common = types[0];
            for (var i = 1; i < types.length && common != null; ++i) {
                var type = types[i];
                if (common !== type) {
                    if (base.isCallableType(common, this.tc) && base.isCallableType(type, this.tc)) {
                        // Fall back to a generic Function type if both types are Function.
                        var fn = ts.createNode(ts.SyntaxKind.FunctionType);
                        fn.parameters = [];
                        var parameter = ts.createNode(ts.SyntaxKind.Parameter);
                        parameter.dotDotDotToken = ts.createNode(ts.SyntaxKind.DotDotDotToken);
                        var name_2 = ts.createNode(ts.SyntaxKind.Identifier);
                        name_2.text = 'args';
                        fn.parameters.push(parameter);
                        common = fn;
                    }
                    else {
                        switch (type.kind) {
                            case ts.SyntaxKind.ArrayType:
                                var array = ts.createNode(ts.SyntaxKind.ArrayType);
                                array.elementType = this.toSimpleDartType([
                                    common.elementType, type.elementType
                                ]);
                                common = array;
                                break;
                            // case ts.SyntaxKind
                            case ts.SyntaxKind.TypeReference:
                                if (common.kind !== ts.SyntaxKind.TypeReference) {
                                    return null;
                                }
                                common = this.commonSupertype(common, type);
                                break;
                            default:
                                return null;
                        }
                    }
                }
            }
            return common;
        }
        return merged;
    };
    FacadeConverter.prototype.toTypeNode = function (type) {
        if (!type)
            return null;
        var symbol = type.getSymbol();
        if (!symbol)
            return null;
        var referenceType = ts.createNode(ts.SyntaxKind.TypeReference);
        // TODO(jacobr): property need to prefix the name better.
        referenceType.typeName = this.createEntityName(symbol);
        referenceType.typeName.parent = referenceType;
        return referenceType;
    };
    FacadeConverter.prototype.createEntityName = function (symbol) {
        var parts = this.tc.getFullyQualifiedName(symbol).split('.');
        var identifier = ts.createNode(ts.SyntaxKind.Identifier);
        identifier.text = parts[parts.length - 1];
        // TODO(jacobr): do we need to include all parts in the entity name?
        return identifier;
    };
    FacadeConverter.prototype.safeGetBaseTypes = function (type) {
        // For an unknown, calling TypeChecker.getBaseTypes on an interface
        // that is a typedef like interface causes the typescript compiler to stack
        // overflow. Not sure if this is a bug in the typescript compiler or I am
        // missing something obvious.
        var declaration = base.getDeclaration(type);
        if (base.isFunctionTypedefLikeInterface(declaration)) {
            return [];
        }
        return this.tc.getBaseTypes(type);
    };
    // TODO(jacobr): all of these subtype checks are fragile and are likely a
    // mistake. We would be better off handling subtype relationships in Dart
    // where we could reuse an existing Dart type system.
    FacadeConverter.prototype.checkTypeSubtypeOf = function (source, target) {
        if (source === target)
            return true;
        if (!(source.flags & ts.TypeFlags.Interface))
            return false;
        var baseTypes = this.safeGetBaseTypes(source);
        for (var i = 0; i < baseTypes.length; ++i) {
            if (baseTypes[i] === target)
                return true;
        }
        return false;
    };
    FacadeConverter.prototype.commonSupertype = function (nodeA, nodeB) {
        if (nodeA == null || nodeB == null)
            return null;
        return this.toTypeNode(this.getCommonSupertype(this.tc.getTypeAtLocation(nodeA), this.tc.getTypeAtLocation(nodeB)));
    };
    FacadeConverter.prototype.getCommonSupertype = function (a, b) {
        if (a === b)
            return a;
        if (!(a.flags & ts.TypeFlags.Interface) || !(b.flags & ts.TypeFlags.Interface)) {
            return null;
        }
        var bestCommonSuperType = null;
        var candidatesA = this.safeGetBaseTypes(a);
        candidatesA.push(a);
        for (var i = 0; i < candidatesA.length; ++i) {
            var type = candidatesA[i];
            if (this.checkTypeSubtypeOf(b, type)) {
                if (!bestCommonSuperType || this.checkTypeSubtypeOf(bestCommonSuperType, type)) {
                    bestCommonSuperType = type;
                }
            }
        }
        return bestCommonSuperType;
    };
    FacadeConverter.prototype.getFileAndName = function (n, originalSymbol) {
        var symbol = originalSymbol;
        while (symbol.flags & ts.SymbolFlags.Alias)
            symbol = this.tc.getAliasedSymbol(symbol);
        var decl = this.getSymbolDeclaration(symbol, n);
        var fileName = decl.getSourceFile().fileName;
        var canonicalFileName = this.getRelativeFileName(fileName)
            .replace(/(\.d)?\.ts$/, '')
            .replace(FACADE_NODE_MODULES_PREFIX, '')
            .replace(this.typingsRootRegex, '');
        var qname = this.tc.getFullyQualifiedName(symbol);
        // Some Qualified Names include their file name. Might be a bug in TypeScript,
        // for the time being just special case.
        if (symbol.flags & (ts.SymbolFlags.Class | ts.SymbolFlags.Function | ts.SymbolFlags.Variable)) {
            qname = symbol.getName();
        }
        if (FACADE_DEBUG)
            console.error('fn:', fileName, 'cfn:', canonicalFileName, 'qn:', qname);
        return { fileName: canonicalFileName, qname: qname };
    };
    // For the Dart keyword list see
    // https://www.dartlang.org/docs/dart-up-and-running/ch02.html#keywords
    FacadeConverter.DART_RESERVED_WORDS = ('assert break case catch class const continue default do else enum extends false final ' +
        'finally for if in is new null rethrow return super switch this throw true try let void ' +
        'while with')
        .split(/ /);
    // These are the built-in and limited keywords.
    FacadeConverter.DART_OTHER_KEYWORDS = ('abstract as async await deferred dynamic export external factory get implements import ' +
        'library operator part set static sync typedef yield call')
        .split(/ /);
    return FacadeConverter;
}(base.TranspilerBase));
exports.FacadeConverter = FacadeConverter;

//# sourceMappingURL=facade_converter.js.map
