"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var ts = require('typescript');
var base = require('./base');
var facade_converter_1 = require('./facade_converter');
var main_1 = require('./main');
var ModuleTranspiler = (function (_super) {
    __extends(ModuleTranspiler, _super);
    function ModuleTranspiler(tr, fc, generateLibraryName) {
        _super.call(this, tr);
        this.fc = fc;
        this.generateLibraryName = generateLibraryName;
    }
    ModuleTranspiler.prototype.visitNode = function (node) {
        switch (node.kind) {
            case ts.SyntaxKind.SourceFile:
                if (this.generateLibraryName) {
                    this.pushContext(main_1.OutputContext.Import);
                    this.emit('@JS()');
                    this.emit('library');
                    this.emit(this.getLibraryName());
                    this.emit(';');
                    this.popContext();
                    this.emitImport('package:js/js.dart');
                }
                ts.forEachChild(node, this.visit.bind(this));
                break;
            case ts.SyntaxKind.EndOfFileToken:
                ts.forEachChild(node, this.visit.bind(this));
                break;
            case ts.SyntaxKind.ImportDeclaration:
                var importDecl = node;
                if (importDecl.importClause) {
                    if (this.isEmptyImport(importDecl))
                        return true;
                    this.emit('import');
                    this.visitExternalModuleReferenceExpr(importDecl.moduleSpecifier);
                    this.visit(importDecl.importClause);
                }
                else {
                    this.reportError(importDecl, 'bare import is unsupported');
                }
                this.emit(';');
                break;
            case ts.SyntaxKind.ImportClause:
                var importClause = node;
                if (importClause.name)
                    this.fc.visitTypeName(importClause.name);
                if (importClause.namedBindings) {
                    this.visit(importClause.namedBindings);
                }
                break;
            case ts.SyntaxKind.NamespaceImport:
                var nsImport = node;
                this.emit('as');
                this.fc.visitTypeName(nsImport.name);
                break;
            case ts.SyntaxKind.NamedImports:
                this.emit('show');
                var used = this.filterImports(node.elements);
                if (used.length === 0) {
                    this.reportError(node, 'internal error, used imports must not be empty');
                }
                this.visitList(used);
                break;
            case ts.SyntaxKind.NamedExports:
                var exportElements = node.elements;
                this.emit('show');
                if (exportElements.length === 0)
                    this.reportError(node, 'empty export list');
                this.visitList(node.elements);
                break;
            case ts.SyntaxKind.ImportSpecifier:
            case ts.SyntaxKind.ExportSpecifier:
                var spec = node;
                if (spec.propertyName) {
                    this.reportError(spec.propertyName, 'import/export renames are unsupported in Dart');
                }
                this.fc.visitTypeName(spec.name);
                break;
            case ts.SyntaxKind.ExportDeclaration:
                var exportDecl = node;
                this.emit('export');
                if (exportDecl.moduleSpecifier) {
                    this.visitExternalModuleReferenceExpr(exportDecl.moduleSpecifier);
                }
                else {
                    this.reportError(node, 're-exports must have a module URL (export x from "./y").');
                }
                if (exportDecl.exportClause)
                    this.visit(exportDecl.exportClause);
                this.emit(';');
                break;
            case ts.SyntaxKind.ImportEqualsDeclaration:
                var importEqDecl = node;
                this.pushContext(main_1.OutputContext.Import);
                this.emit('import');
                this.visit(importEqDecl.moduleReference);
                this.emit('as');
                this.fc.visitTypeName(importEqDecl.name);
                this.emit(';');
                this.popContext();
                break;
            case ts.SyntaxKind.ExternalModuleReference:
                this.visitExternalModuleReferenceExpr(node.expression);
                break;
            default:
                return false;
        }
        return true;
    };
    ModuleTranspiler.isIgnoredImport = function (e) { return false; };
    ModuleTranspiler.prototype.visitExternalModuleReferenceExpr = function (expr) {
        // TODO: what if this isn't a string literal?
        var moduleName = expr;
        var text = moduleName.text;
        if (text.match(/^\.\//)) {
            // Strip './' to be more Dart-idiomatic.
            text = text.substring(2);
        }
        else if (!text.match(/^\.\.\//)) {
            // Unprefixed imports are package imports.
            text = 'package:' + text;
        }
        this.emit(JSON.stringify(text + '.dart'));
    };
    ModuleTranspiler.prototype.isEmptyImport = function (n) {
        var bindings = n.importClause.namedBindings;
        if (bindings.kind !== ts.SyntaxKind.NamedImports)
            return false;
        var elements = bindings.elements;
        if (elements.length === 0)
            return true;
        return elements.every(ModuleTranspiler.isIgnoredImport);
    };
    ModuleTranspiler.prototype.filterImports = function (ns) {
        return ns.filter(function (e) { return !ModuleTranspiler.isIgnoredImport(e); });
    };
    ModuleTranspiler.prototype.getLibraryName = function (nameForTest) {
        var fileName = this.getRelativeFileName(nameForTest);
        var parts = fileName.split('/');
        return parts.filter(function (p) { return p.length > 0; })
            .map(function (p) { return p.replace(/[^\w.]/g, '_'); })
            .map(function (p) { return p.replace(/\.d\.ts$/g, ''); })
            .map(function (p) { return p.replace(/\.[jt]s$/g, ''); })
            .map(function (p) { return facade_converter_1.FacadeConverter.DART_RESERVED_WORDS.indexOf(p) !== -1 ? '_' + p : p; })
            .join('.');
    };
    return ModuleTranspiler;
}(base.TranspilerBase));
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = ModuleTranspiler;

//# sourceMappingURL=module.js.map
