"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var ts = require('typescript');
var base = require('./base');
var facade_converter_1 = require('./facade_converter');
var TypeTranspiler = (function (_super) {
    __extends(TypeTranspiler, _super);
    function TypeTranspiler(tr, fc) {
        _super.call(this, tr);
        this.fc = fc;
    }
    TypeTranspiler.prototype.visitNode = function (node) {
        if (base.isTypeNode(node)) {
            this.emit(this.fc.generateDartTypeName(node));
            return true;
        }
        switch (node.kind) {
            case ts.SyntaxKind.TypeAssertionExpression:
                var typeAssertExpr = node;
                if (this.isReifiedTypeLiteral(typeAssertExpr)) {
                    this.visit(typeAssertExpr.expression);
                    break; // type is handled by the container literal itself.
                }
                this.emit('(');
                this.visit(typeAssertExpr.expression);
                this.emit('as');
                this.visit(typeAssertExpr.type);
                this.emit(')');
                break;
            case ts.SyntaxKind.TypeParameter:
                var typeParam = node;
                this.visit(typeParam.name);
                if (typeParam.constraint) {
                    this.emit('extends');
                    this.visit(typeParam.constraint);
                }
                break;
            case ts.SyntaxKind.QualifiedName:
                // TODO(jacobr): there is overlap between this case and
                // generateDartTypeName in facade_converter.
                var first = node;
                var match = this.fc.lookupCustomDartTypeName(first);
                if (match) {
                    this.emit(match.name);
                    if (match.comment) {
                        this.emit('/* ' + match.comment + '*/');
                    }
                    break;
                }
                this.visit(first.left);
                this.emit('.');
                this.visit(first.right);
                break;
            case ts.SyntaxKind.Identifier:
                var ident = node;
                var text = facade_converter_1.fixupIdentifierName(ident.text);
                this.emit(text);
                break;
            // TODO(jacobr): all these cases might be obsolete.
            case ts.SyntaxKind.NumberKeyword:
                this.emit('num');
                break;
            case ts.SyntaxKind.StringKeyword:
                this.emit('String');
                break;
            case ts.SyntaxKind.VoidKeyword:
                this.emit('void');
                break;
            case ts.SyntaxKind.BooleanKeyword:
                this.emit('bool');
                break;
            case ts.SyntaxKind.AnyKeyword:
                this.emit('dynamic');
                break;
            default:
                return false;
        }
        return true;
    };
    TypeTranspiler.prototype.isReifiedTypeLiteral = function (node) {
        if (node.expression.kind === ts.SyntaxKind.ArrayLiteralExpression &&
            node.type.kind === ts.SyntaxKind.ArrayType) {
            return true;
        }
        else if (node.expression.kind === ts.SyntaxKind.ObjectLiteralExpression &&
            node.type.kind === ts.SyntaxKind.TypeLiteral) {
            return true;
        }
        return false;
    };
    return TypeTranspiler;
}(base.TranspilerBase));
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = TypeTranspiler;

//# sourceMappingURL=type.js.map
